/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004-2005 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License ONLY.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <glib.h>

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>

#define EXO_API_SUBJECT_TO_CHANGE
#include <exo/exo.h>

#include "xfmedia-xine.h"
#include "mainwin.h"
#include <xfmedia/xfmedia-video-window.h>
#include <xfmedia/xfmedia-settings.h>
#include "trayicon.h"
#include "keybindings.h"
#include "main.h"
#include "xfmedia-internal.h"
#include "mediamarks.h"
#include "http.h"
#include "playlist-files.h"
/*#include "equaliser.h"*/

#define MAX_SHUFFLE_QUEUE_DEPTH 100
#define AUTOHIDE_TIMEOUT        500


gboolean
xfmedia_mainwin_check_do_autohide(XfmediaMainwin *mwin)
{
    gboolean do_it = FALSE;
    GdkScreen *gscreen;
    gint scr_w, scr_h, x, y, w, h, ptr_x, ptr_y, a_x, a_y, a_w, a_h;
    GdkModifierType modmask;
    GdkRectangle extents;
    
    if(!mwin->autohide_enabled)
        return FALSE;
    
    gscreen = gtk_widget_get_screen(mwin->window);
    scr_w = gdk_screen_get_width(gscreen);
    scr_h = gdk_screen_get_height(gscreen);
    
    gdk_window_get_frame_extents(mwin->window->window, &extents);
    x = extents.x;
    y = extents.y;
    w = extents.width;
    h = extents.height;
    
    gdk_window_get_pointer(gdk_screen_get_root_window(gscreen), &ptr_x, &ptr_y,
            &modmask);
    if(xfmedia_rectangle_contains_point(&extents, ptr_x, ptr_y))
        return FALSE;
    
    if(y == 0) {
        a_x = x;
        a_y = 0;
        a_w = w;
        a_h = 3;
        do_it = TRUE;
    } else if(y == scr_h - h) {
        a_x = x;
        a_y = scr_h - 3;
        a_w = w;
        a_h = 3;
        do_it = TRUE;
    } else if(x == 0) {
        a_x = 0;
        a_y = y;
        a_w = 3;
        a_h = h;
        do_it = TRUE;
    } else if(x == scr_w - w) {
        a_x = scr_w - 3;
        a_y = y;
        a_w = 3;
        a_h = h;
        do_it = TRUE;
    }
    
    if(do_it && !mwin->volume_window) {
        gtk_widget_set_size_request(mwin->autohide_window, a_w, a_h);
        gtk_widget_realize(mwin->autohide_window);
        gtk_window_resize(GTK_WINDOW(mwin->autohide_window), a_w, a_h);
        gtk_window_move(GTK_WINDOW(mwin->autohide_window), a_x, a_y);
        if(xfmedia_settings_get_bool("/xfmedia/general/window_sticky"))
            gtk_window_stick(GTK_WINDOW(mwin->autohide_window));
        else
            gtk_window_unstick(GTK_WINDOW(mwin->autohide_window));
        gtk_window_set_keep_above(GTK_WINDOW(mwin->autohide_window), TRUE);
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(mwin->autohide_window), TRUE);
        gtk_window_set_accept_focus(GTK_WINDOW(mwin->autohide_window), FALSE);
        gtk_widget_show(mwin->autohide_window);
        gtk_widget_hide(mwin->window);
        
        if(xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window)))
            gdk_window_lower(mwin->autohide_window->window);
    }
    
    mwin->autohide_timer = 0;
    
    return FALSE;
}

void
xfmedia_mainwin_set_autohide(XfmediaMainwin *mwin, gboolean do_autohide)
{
    g_return_if_fail(mwin);
    
    if(mwin->autohide_enabled == do_autohide)
        return;
    
    mwin->autohide_enabled = do_autohide;
    
    if(!do_autohide)
        xfmedia_autohidewin_enter_notify(NULL, NULL, mwin);
    else
        mwin->autohide_timer = g_timeout_add(AUTOHIDE_TIMEOUT,
                (GSourceFunc)xfmedia_mainwin_check_do_autohide, mwin);
}

static void
instantiate_extra_xine(XfmediaMainwin *mwin)
{
    xine_t *xine;
    xine_cfg_entry_t cfgentry_a, cfgentry_v;
    gint real_aud_bufsize = -1, real_vid_bufsize = -1;
    
    xine = xfmedia_xine_get_raw_xine_engine(mwin->xfx);
    g_return_if_fail(xine);
    
    /* save buffer sizes */
    if(xine_config_lookup_entry(xine, "video.num_buffers", &cfgentry_v)
       && cfgentry_v.type == XINE_CONFIG_TYPE_NUM)
    {
        real_vid_bufsize = cfgentry_v.num_value;
        cfgentry_v.num_value = 5;
        xine_config_update_entry(xine, &cfgentry_v);
    }
    if(xine_config_lookup_entry(xine, "audio.num_buffers", &cfgentry_a)
       && cfgentry_a.type == XINE_CONFIG_TYPE_NUM)
    {
        real_aud_bufsize = cfgentry_a.num_value;
        cfgentry_a.num_value = 5;
        xine_config_update_entry(xine, &cfgentry_a);
    }
    
    mwin->extra_xine_mx = g_mutex_new();
    mwin->extra_vp = xine_open_video_driver(xine, "none",
                                            XINE_VISUAL_TYPE_NONE, NULL);
    mwin->extra_ap = xine_open_audio_driver(xine, "none", NULL);
    mwin->extra_stream = xine_stream_new(xine, mwin->extra_ap, mwin->extra_vp);
    
    /* restore buffer sizes */
    if(real_vid_bufsize != -1) {
        cfgentry_v.num_value = real_vid_bufsize;
        xine_config_update_entry(xine, &cfgentry_v);
    }
    if(real_aud_bufsize != -1) {
        cfgentry_v.num_value = real_aud_bufsize;
        xine_config_update_entry(xine, &cfgentry_a);
    }
}

static void
create_autohide_window(XfmediaMainwin *mwin)
{
    GtkWidget *win;
    guint opacity;
    gint w, h;
    
    win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(win), "xfmedia-autohide-window");
    gtk_window_set_decorated(GTK_WINDOW(win), FALSE);
    gtk_window_set_keep_above(GTK_WINDOW(win), TRUE);
    gtk_window_set_skip_taskbar_hint(GTK_WINDOW(win), TRUE);
    gtk_window_set_skip_pager_hint(GTK_WINDOW(win), TRUE);
    g_signal_connect(G_OBJECT(win), "enter-notify-event",
            G_CALLBACK(xfmedia_autohidewin_enter_notify), mwin);
    
    gtk_window_get_size(GTK_WINDOW(mwin->window), &w, &h);
    gtk_widget_set_size_request(win, w, h);
    
    /* set 40% transparency (if supported) */
    opacity = (guint)(0.6*0xffffffff);
    gdk_property_change(mwin->fs_controls->window,
            gdk_atom_intern("_NET_WM_WINDOW_OPACITY", FALSE),
            gdk_atom_intern("CARDINAL", FALSE), 32,
            GDK_PROP_MODE_REPLACE, (const guchar *)&opacity, sizeof(guint));
    
    mwin->autohide_window = win;
}

static void
mainwin_create_progress_box(XfmediaMainwin *mwin)
{
    PangoFontDescription *fdesc;
    gint size;
    
    mwin->progress_box = gtk_hbox_new(FALSE, BORDER/2);
    gtk_container_set_border_width(GTK_CONTAINER(mwin->progress_box), 0);
    
    mwin->progress_label = gtk_label_new("");
    gtk_widget_show(mwin->progress_label);
    gtk_box_pack_start(GTK_BOX(mwin->progress_box), mwin->progress_label,
            FALSE, FALSE, 0);
    
    fdesc = pango_font_description_copy(mwin->progress_label->style->font_desc);
    pango_font_description_set_style(fdesc, PANGO_STYLE_ITALIC);
    size = pango_font_description_get_size(fdesc);
    pango_font_description_set_size(fdesc, size*PANGO_SCALE_LARGE);
    gtk_widget_modify_font(mwin->progress_label, fdesc);
    pango_font_description_free(fdesc);
    
    mwin->progress_bar = gtk_progress_bar_new();
    gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(mwin->progress_bar),
            GTK_PROGRESS_LEFT_TO_RIGHT);
    gtk_widget_show(mwin->progress_bar);
    gtk_box_pack_start(GTK_BOX(mwin->progress_box), mwin->progress_bar,
            TRUE, TRUE, 0);
}

static void
mainwin_create_ui_msg_label(XfmediaMainwin *mwin)
{
    PangoFontDescription *fdesc;
    
    mwin->ui_message_label = gtk_label_new("");
    gtk_label_set_ellipsize(GTK_LABEL(mwin->ui_message_label),
                            PANGO_ELLIPSIZE_END);
    gtk_misc_set_alignment(GTK_MISC(mwin->ui_message_label), 0.0, 0.5);
    
    fdesc = pango_font_description_from_string("italic");
    gtk_widget_modify_font(mwin->ui_message_label, fdesc);
    pango_font_description_free(fdesc);
}

static void
create_fs_control_window(XfmediaMainwin *mwin)
{
    GtkWidget *tophbox, *img, *btn, *lbl, *slider, *evtbox;
    gint iw, ih;
    guint opacity;
    
    mwin->fs_controls = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_window_set_title(GTK_WINDOW(mwin->fs_controls),
            "xfmedia-fullscreen-control-window");
    gtk_window_set_decorated(GTK_WINDOW(mwin->fs_controls), FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(mwin->fs_controls), BORDER/2);
    
    tophbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(tophbox);
    gtk_container_add(GTK_CONTAINER(mwin->fs_controls), tophbox);
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_BUTTON, &iw, &ih);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PREVIOUS,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(tophbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_prev_cb), mwin);
    
    mwin->fs_play_btn_img = img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PLAY,
            GTK_ICON_SIZE_BUTTON);
    g_object_ref(G_OBJECT(img));
    gtk_widget_show(img);
    mwin->fs_playpause_btn = btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), mwin->fs_play_btn_img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(tophbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playpause_cb), mwin);
    
    mwin->fs_pause_btn_img = img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PAUSE,
            GTK_ICON_SIZE_BUTTON);
    g_object_ref(G_OBJECT(img));
    gtk_widget_show(img);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_STOP,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(tophbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_stop_cb), mwin);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_NEXT,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(tophbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_next_cb), mwin);
    
    evtbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evtbox), FALSE);
    gtk_widget_show(evtbox);
    gtk_box_pack_start(GTK_BOX(tophbox), evtbox, FALSE, FALSE, 0);
    gtk_widget_add_events(evtbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(evtbox), "button-press-event",
            G_CALLBACK(xfmedia_pos_lbl_click_cb), mwin);
    
    mwin->fs_time_label = lbl = gtk_label_new(NULL);
    gtk_widget_show(lbl);
    gtk_container_add(GTK_CONTAINER(evtbox), lbl);
    
    mwin->fs_position_slider = slider = gtk_hscale_new(NULL);
    gtk_scale_set_draw_value(GTK_SCALE(slider), FALSE);
    gtk_range_set_increments(GTK_RANGE(slider), 1, 10);
    gtk_range_set_update_policy(GTK_RANGE(slider), GTK_UPDATE_DISCONTINUOUS);
    gtk_widget_show(slider);
    gtk_box_pack_start(GTK_BOX(tophbox), slider, TRUE, TRUE, 0);
    gtk_widget_add_events(mwin->fs_position_slider, GDK_MOTION_NOTIFY);
    g_signal_connect(G_OBJECT(slider), "value-changed",
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
    g_signal_connect(G_OBJECT(slider), "button-press-event",
            G_CALLBACK(xfmedia_pos_slider_btn_press_cb), mwin);
    g_signal_connect(G_OBJECT(slider), "button-release-event",
            G_CALLBACK(xfmedia_pos_slider_btn_rls_cb), mwin);
    
    gtk_widget_realize(mwin->fs_controls);
    
    /* set 40% transparency (if supported) */
    opacity = (guint)(0.6*0xffffffff);
    gdk_property_change(mwin->fs_controls->window,
            gdk_atom_intern("_NET_WM_WINDOW_OPACITY", FALSE),
            gdk_atom_intern("CARDINAL", FALSE), 32,
            GDK_PROP_MODE_REPLACE, (const guchar *)&opacity, sizeof(guint));
}

static GtkWidget *
create_autoplay_menu_item(XfmediaMainwin *mwin)
{
    GtkWidget *menu, *ap_mi = NULL, *mi, *img;
    GList *autoplay_plugins, *l;
    
    /* xine's autoplay menu */
    autoplay_plugins = xfmedia_xine_get_autoplay_input_plugin_ids(mwin->xfx);
    if(autoplay_plugins) {
        img = gtk_image_new_from_stock(GTK_STOCK_NEW, GTK_ICON_SIZE_MENU);
        gtk_widget_show(img);
        ap_mi = mi = gtk_image_menu_item_new_with_mnemonic(_("_Other"));
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
        gtk_widget_show(mi);
        
        menu = gtk_menu_new();
        gtk_widget_show(menu);
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), menu);
        
        for(l = autoplay_plugins; l; l = l->next) {
            mi = gtk_image_menu_item_new_with_label((const gchar *)l->data);
            gtk_widget_show(mi);
            if(strstr((const gchar *)l->data, "DVD") == l->data
                    || strstr((const gchar *)l->data, "CD") == l->data
                    || strstr((const gchar *)l->data, "VCD") == l->data)
            {
                img = gtk_image_new_from_stock(GTK_STOCK_CDROM,
                        GTK_ICON_SIZE_MENU);
                gtk_widget_show(img);
                gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
            }
            gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
            g_signal_connect(G_OBJECT(mi), "activate",
                    G_CALLBACK(xfmedia_mainwin_playlist_add_autoplay_cb), mwin);
        }
        g_list_free(autoplay_plugins);
    }
    
    return ap_mi;
}

static GtkWidget *
create_add_menu(XfmediaMainwin *mwin)
{
    GtkWidget *menu, *mi, *img;
    
    mwin->add_menu = menu = gtk_menu_new();
    gtk_widget_show(menu);
    
    mi = create_autoplay_menu_item(mwin);
    if(mi) {
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    }
    
    img = gtk_image_new_from_stock(GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    mi = gtk_image_menu_item_new_with_mnemonic(_("_URL..."));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_add_url_cb), mwin);
    
    img = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Directory..."));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_add_dir_cb), mwin);
    
    img = gtk_image_new_from_stock(GTK_STOCK_HARDDISK, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    mi = gtk_image_menu_item_new_with_mnemonic(_("_File..."));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_add_file_cb), mwin);
    
    return menu;
}

XfmediaMainwin *
xfmedia_mainwin_new(XfmediaXine *xfx, GtkWindow *video_window)
{
    XfmediaMainwin *mwin;
    GtkWidget *topwin, *topvbox, *playlist_pane, *menu2;
    GtkWidget *hbox, *slider, *vbox, *btn, *lbl, *img, *menu, *mi, *evtbox;
    gint i, iw, ih, w, h_s, h_l;
    GtkTooltips *ttips;
    gchar *user_kb_file, *extensions, **ext_array;
    GString *ext_gstr;
    
    g_return_val_if_fail(xfx != NULL, NULL);
    
    mwin = g_new0(XfmediaMainwin, 1);
    mwin->session_id = -1;
    
    DBG("setting up xfx signals");
    mwin->xfx = xfx;
    g_signal_connect(G_OBJECT(xfx), "button-press-event",
            G_CALLBACK(xfmedia_vwin_button_press_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "stream-ended",
            G_CALLBACK(xfmedia_xfmedia_xine_stream_ended_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "ui-message",
            G_CALLBACK(xfmedia_xfmedia_xine_ui_message_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "progress-message",
            G_CALLBACK(xfmedia_xfmedia_xine_progress_message_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "format-changed",
            G_CALLBACK(xfmedia_xfmedia_xine_format_changed_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "mrl-reference",
            G_CALLBACK(xfmedia_xfmedia_xine_mrl_reference_cb), mwin);
    g_signal_connect(G_OBJECT(xfx), "set-title",
            G_CALLBACK(xfmedia_xfmedia_xine_set_title_cb), mwin);
    DBG("  done");
    
    instantiate_extra_xine(mwin);
    
    mwin->video_window = GTK_WIDGET(video_window);
    gtk_widget_add_events(GTK_WIDGET(video_window), GDK_BUTTON_PRESS_MASK|GDK_KEY_RELEASE_MASK);
    g_signal_connect(G_OBJECT(video_window), "button-press-event",
            G_CALLBACK(xfmedia_vwin_button_press_cb), mwin);
    g_signal_connect(G_OBJECT(video_window), "motion-notify-event",
            G_CALLBACK(xfmedia_vwin_motion_notify_cb), mwin);
    g_signal_connect(G_OBJECT(video_window), "delete-event",
            G_CALLBACK(xfmedia_vwin_delete_cb), mwin);
    g_signal_connect(G_OBJECT(video_window), "key-release-event",
            G_CALLBACK(xfmedia_mainwin_keyrelease_cb), mwin);
    g_signal_connect(G_OBJECT(video_window), "map-event",
            G_CALLBACK(xfmedia_vwin_mapunmap_cb), mwin);
    g_signal_connect(G_OBJECT(video_window), "unmap-event",
            G_CALLBACK(xfmedia_vwin_mapunmap_cb), mwin);
    
    mwin->tooltips = ttips = gtk_tooltips_new();
    
    mwin->window = topwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(topwin), "Xfmedia");
    gtk_window_set_role(GTK_WINDOW(topwin), "control/playlist window");
    gtk_container_set_border_width(GTK_CONTAINER(topwin), BORDER/2);
    if(xfmedia_settings_get_bool("/xfmedia/playlist/playlist_expanded")) {
        w = xfmedia_settings_get_int("/xfmedia/general/width");
        h_l = xfmedia_settings_get_int("/xfmedia/general/height_large");
        if(h_l != -1 &&  w != -1)
            gtk_window_resize(GTK_WINDOW(topwin), w, h_l);
        else
            gtk_window_resize(GTK_WINDOW(topwin), 475, 550);
    } else {
        w = xfmedia_settings_get_int("/xfmedia/general/width");
        h_s = xfmedia_settings_get_int("/xfmedia/general/height_small");
        if(h_s != -1 && w != -1)
            gtk_window_resize(GTK_WINDOW(topwin), w, h_s);
        else {
            gtk_window_get_size(GTK_WINDOW(topwin), &w, &h_s);
            gtk_window_resize(GTK_WINDOW(topwin), 500, h_s);
        }
    }
    gtk_widget_add_events(topwin, GDK_BUTTON_PRESS_MASK|GDK_STRUCTURE_MASK
            |GDK_SUBSTRUCTURE_MASK|GDK_KEY_RELEASE_MASK);
    g_signal_connect(G_OBJECT(topwin), "button-release-event",
            G_CALLBACK(xfmedia_mainwin_menu_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "configure-event",
            G_CALLBACK(xfmedia_mainwin_configure_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "window-state-event",
            G_CALLBACK(xfmedia_mainwin_winstate_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "leave-notify-event",
            G_CALLBACK(xfmedia_mainwin_enterleave_notify_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "enter-notify-event",
            G_CALLBACK(xfmedia_mainwin_enterleave_notify_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "map-event",
            G_CALLBACK(xfmedia_mainwin_map_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "unmap-event",
            G_CALLBACK(xfmedia_mainwin_unmap_cb), mwin);
    g_signal_connect(G_OBJECT(topwin), "key-release-event",
            G_CALLBACK(xfmedia_mainwin_keyrelease_cb), mwin);
    
    xfmedia_keybindings_init();
    user_kb_file = xfce_resource_save_location(XFCE_RESOURCE_CONFIG,
            "xfmedia/keybindings.rc", FALSE);
    if(user_kb_file && g_file_test(user_kb_file, G_FILE_TEST_EXISTS)) {
        xfmedia_keybindings_load(SYSCONFDIR "/xdg/xfmedia/keybindings.rc");
        xfmedia_keybindings_load(user_kb_file);
    } else
        xfmedia_keybindings_load(NULL);
    if(user_kb_file)
        g_free(user_kb_file);
    
    topvbox = gtk_vbox_new(FALSE, 2);
    gtk_widget_show(topvbox);
    gtk_container_add(GTK_CONTAINER(topwin), topvbox);
    gtk_widget_add_events(topvbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(topvbox), "button-press-event",
            G_CALLBACK(xfmedia_mainwin_menu_cb), mwin);
    
    evtbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evtbox), FALSE);
    gtk_widget_show(evtbox);
    gtk_box_pack_start(GTK_BOX(topvbox), evtbox, FALSE, FALSE, 0);
    gtk_widget_add_events(evtbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(evtbox), "button-press-event",
            G_CALLBACK(xfmedia_mainwin_menu_cb), mwin);
            
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(evtbox), hbox);
    
    evtbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evtbox), FALSE);
    gtk_widget_show(evtbox);
    gtk_box_pack_start(GTK_BOX(hbox), evtbox, FALSE, FALSE, 0);
    gtk_widget_add_events(evtbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(evtbox), "button-press-event",
            G_CALLBACK(xfmedia_pos_lbl_click_cb), mwin);
    
    mwin->time_label = lbl = gtk_label_new(NULL);
    gtk_widget_show(lbl);
    gtk_container_add(GTK_CONTAINER(evtbox), lbl);
    
    mwin->song_label = lbl = gtk_label_new("");
    gtk_label_set_ellipsize(GTK_LABEL(lbl), PANGO_ELLIPSIZE_END);
    gtk_label_set_markup(GTK_LABEL(lbl), "<span size=\"large\"> </span>");
    gtk_label_set_use_markup(GTK_LABEL(lbl), TRUE);
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, TRUE, TRUE, 0);
    
    mainwin_create_progress_box(mwin);
    gtk_box_pack_start(GTK_BOX(hbox), mwin->progress_box, TRUE, TRUE, 0);
    
    mainwin_create_ui_msg_label(mwin);
    gtk_box_pack_start(GTK_BOX(hbox), mwin->ui_message_label, TRUE, TRUE, 0);
    
    evtbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evtbox), FALSE);
    gtk_widget_show(evtbox);
    gtk_box_pack_start(GTK_BOX(topvbox), evtbox, FALSE, FALSE, 0);
    gtk_widget_add_events(evtbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(evtbox), "button-press-event",
            G_CALLBACK(xfmedia_mainwin_menu_cb), mwin);
    
    hbox = gtk_hbox_new(FALSE, BORDER-2);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(evtbox), hbox);
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_BUTTON, &iw, &ih);
    
    img = xfmedia_smaller_stock_icon_load(XFMEDIA_STOCK_VOLUME,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    mwin->volume_btn = btn = gtk_toggle_button_new();
    gtk_widget_add_events(btn, GDK_BUTTON_PRESS);
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "toggled",
            G_CALLBACK(xfmedia_mainwin_show_volume_window_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Volume"), NULL);
    
    mwin->position_slider = slider = gtk_hscale_new(NULL);
    gtk_scale_set_draw_value(GTK_SCALE(slider), FALSE);
    gtk_range_set_increments(GTK_RANGE(slider), 1, 10);
    gtk_range_set_update_policy(GTK_RANGE(slider), GTK_UPDATE_DISCONTINUOUS);
    gtk_widget_show(slider);
    gtk_box_pack_start(GTK_BOX(hbox), slider, TRUE, TRUE, 0);
    gtk_widget_add_events(mwin->position_slider, GDK_MOTION_NOTIFY);
    g_signal_connect(G_OBJECT(slider), "value-changed",
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
    g_signal_connect(G_OBJECT(slider), "button-press-event",
            G_CALLBACK(xfmedia_pos_slider_btn_press_cb), mwin);
    g_signal_connect(G_OBJECT(slider), "button-release-event",
            G_CALLBACK(xfmedia_pos_slider_btn_rls_cb), mwin);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PREVIOUS,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_prev_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Previous Track"), NULL);
    
    mwin->play_btn_img = img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PLAY,
            GTK_ICON_SIZE_BUTTON);
    g_object_ref(G_OBJECT(img));
    gtk_widget_show(img);
    mwin->playpause_btn = btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playpause_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Play/Pause Track"), NULL);
    
    mwin->pause_btn_img = img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_PAUSE,
            GTK_ICON_SIZE_BUTTON);
    g_object_ref(G_OBJECT(img));
    gtk_widget_show(img);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_STOP,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_stop_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Stop Playback"), NULL);
    
    img = xfmedia_squished_stock_icon_load(GTK_STOCK_MEDIA_NEXT,
            GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_next_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Next Track"), NULL);
    
    vbox = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox);
    hbox = gtk_hbox_new(FALSE, BORDER-4);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    
    mwin->playlist_pane = playlist_pane = gtk_expander_new_with_mnemonic(_("_Playlist"));
    gtk_expander_set_expanded(GTK_EXPANDER(playlist_pane),
            xfmedia_settings_get_bool("/xfmedia/playlist/playlist_expanded"));
    gtk_widget_show(playlist_pane);
    gtk_box_pack_start(GTK_BOX(topvbox), playlist_pane, TRUE, TRUE, 0);
    g_signal_connect_after(G_OBJECT(playlist_pane), "activate",
            G_CALLBACK(xfmedia_mainwin_expander_activate_cb), mwin);
    
    mwin->plist = XFMEDIA_PLAYLIST(xfmedia_playlist_new());
    xfmedia_playlist_set_shuffle_state(mwin->plist,
            xfmedia_settings_get_bool("/xfmedia/playlist/shuffle"));
    xfmedia_playlist_set_repeat_state(mwin->plist,
           xfmedia_settings_get_bool("/xfmedia/playlist/repeat"));
    gtk_widget_show(GTK_WIDGET(mwin->plist));
    gtk_container_add(GTK_CONTAINER(playlist_pane), GTK_WIDGET(mwin->plist));
    g_signal_connect(G_OBJECT(mwin->plist), "playlist-scrolled",
            G_CALLBACK(xfmedia_mainwin_playlist_scrolled_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "entry-activated",
            G_CALLBACK(xfmedia_mainwin_playlist_entry_activated_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "shuffle-toggled",
            G_CALLBACK(xfmedia_mainwin_playlist_shuffle_toggled_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "repeat-toggled",
            G_CALLBACK(xfmedia_mainwin_playlist_repeat_toggled_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "playlist-cleared",
            G_CALLBACK(xfmedia_mainwin_playlist_cleared_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "playlist-opened",
            G_CALLBACK(xfmedia_mainwin_playlist_opened_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "playlist-saved",
            G_CALLBACK(xfmedia_mainwin_playlist_saved_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "entry-added",
            G_CALLBACK(xfmedia_mainwin_playlist_entry_added_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "entry-changed",
            G_CALLBACK(xfmedia_mainwin_playlist_entry_changed_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "entry-removed",
            G_CALLBACK(xfmedia_mainwin_playlist_entry_removed_cb), mwin);
    g_signal_connect(G_OBJECT(mwin->plist), "add-button-clicked",
            G_CALLBACK(xfmedia_mainwin_playlist_add_button_cb), mwin);
    
    extensions = xfmedia_xine_get_file_extensions(mwin->xfx);
    ext_gstr = g_string_sized_new(strlen(extensions)+35);
    ext_array = g_strsplit(extensions, " ", 0);
    g_free(extensions);
    for(i = 0; ext_array[i]; i++)
        g_string_append_printf(ext_gstr, ".%s ", ext_array[i]);
    g_strfreev(ext_array);
    xfmedia_playlist_set_allowed_extensions(mwin->plist, ext_gstr->str);
    g_string_free(ext_gstr, TRUE);
    
    mwin->add_menu = create_add_menu(mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    xfmedia_playlist_context_menu_prepend_item(mwin->plist, GTK_MENU_ITEM(mi));
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Enqueue"));
    img = gtk_image_new_from_stock(GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    xfmedia_playlist_context_menu_prepend_item(mwin->plist, GTK_MENU_ITEM(mi));
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
                             G_CALLBACK(xfmedia_mainwin_queue_selected), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("Add _Mediamark..."));
    img = gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    xfmedia_playlist_context_menu_prepend_item(mwin->plist, GTK_MENU_ITEM(mi));
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_playlist_add_mm_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    xfmedia_playlist_context_menu_prepend_item(mwin->plist, GTK_MENU_ITEM(mi));
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_File Info"));
    img = gtk_image_new_from_stock(GTK_STOCK_DIALOG_INFO, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    xfmedia_playlist_context_menu_prepend_item(mwin->plist, GTK_MENU_ITEM(mi));
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_playlist_file_info_cb), mwin);
    
    /* right-click popup menu */
    
    mwin->popup_menu = menu = gtk_menu_new();
    gtk_widget_show(menu);
    g_signal_connect_data(G_OBJECT(menu), "deactivate",
                           G_CALLBACK(xfmedia_mainwin_enable_autohide), mwin,
                           NULL, G_CONNECT_AFTER|G_CONNECT_SWAPPED);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("About _Xfmedia"));
    img = gtk_image_new_from_stock(GTK_STOCK_ABOUT, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_about_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Preferences"));
    img = gtk_image_new_from_stock(GTK_STOCK_PREFERENCES, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_settings_show_dialog), mwin);
    
#if 0
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Equalizer..."));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_equaliser_show), mwin);
#endif

    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    mi = gtk_image_menu_item_new_with_mnemonic(_("_Jump to File..."));
    img = gtk_image_new_from_stock(GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_trigger_jump_to_file), mwin->plist);
    
    mi = gtk_check_menu_item_new_with_mnemonic(_("Play_list Queue..."));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
                     G_CALLBACK(xfmedia_mainwin_toggle_playlist_queue), mwin);
    mwin->playlist_q_mi = mi;
                     
    mwin->mmarks_mi = mi = gtk_image_menu_item_new_with_mnemonic(_("_Mediamarks"));
    img = gtk_image_new_from_stock(GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    menu2 = xfmedia_mediamarks_create_menu(mwin);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), menu2);
    xfmedia_mediamarks_hook_changed(xfmedia_mainwin_mediamarks_changed_cb, mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("Re_order"));
    img = gtk_image_new_from_stock(GTK_STOCK_SORT_ASCENDING, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    menu2 = gtk_menu_new();
    gtk_widget_show(menu2);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), menu2);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("Sort by _Title"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu2), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_sort_title_cb), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("Sort by _Filename"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu2), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_sort_filename_cb), mwin);
            
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Randomize List"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu2), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_randomise_playlist_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &iw, &ih);
    
    mwin->shuffle_mi = mi = gtk_check_menu_item_new_with_mnemonic(_("_Shuffle"));
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
            xfmedia_settings_get_bool("/xfmedia/playlist/shuffle"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "toggled",
            G_CALLBACK(xfmedia_shuffle_menu_toggle_cb), mwin);
    
    mwin->repeat_mi = mi = gtk_check_menu_item_new_with_mnemonic(_("_Repeat"));
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
            xfmedia_settings_get_bool("/xfmedia/playlist/repeat"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "toggled",
            G_CALLBACK(xfmedia_repeat_menu_toggle_cb), mwin);
    
    mi = gtk_check_menu_item_new_with_mnemonic(_("Repeat Single _Track"));
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
            xfmedia_settings_get_bool("/xfmedia/playlist/repeat_one"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "toggled",
            G_CALLBACK(xfmedia_repeat_one_menu_toggle_cb), mwin);
    
    mwin->svis_mi = mi = gtk_check_menu_item_new_with_mnemonic(_("Show Audio _Visualization"));
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
            xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis"));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "toggled",
            G_CALLBACK(xfmedia_show_vis_menu_toggle_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    mi = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_mainwin_quit_mi_cb), mwin);
    
    gtk_widget_realize(topwin);
    
    create_fs_control_window(mwin);
    create_autohide_window(mwin);
    
    mwin->playlist_q = XFMEDIA_PLAYLIST_QUEUE(xfmedia_playlist_queue_new(GTK_WINDOW(mwin->window)));
    gtk_widget_add_events(GTK_WIDGET(mwin->playlist_q), GDK_KEY_RELEASE
                                                        |GDK_SUBSTRUCTURE_MASK);
    g_signal_connect(G_OBJECT(mwin->playlist_q), "key-release-event",
                     G_CALLBACK(xfmedia_mainwin_keyrelease_cb), mwin);
    g_signal_connect_swapped(G_OBJECT(mwin->playlist_q), "map-event",
                             G_CALLBACK(xfmedia_mainwin_disable_autohide), mwin);
    g_signal_connect_swapped(G_OBJECT(mwin->playlist_q), "unmap-event",
                             G_CALLBACK(xfmedia_mainwin_enable_autohide), mwin);
    g_signal_connect(G_OBJECT(mwin->playlist_q), "configure-event",
                     G_CALLBACK(xfmedia_mainwin_playlist_q_configure_cb), mwin);
    
    xfmedia_mainwin_set_time_label(mwin, 0, FALSE, NULL);
    
    mwin->shuffle_history = g_queue_new();
    
    h_s = xfmedia_settings_get_int("/xfmedia/general/height_small");
    if(xfmedia_settings_get_bool("/xfmedia/playlist/playlist_expanded")) {
        h_l = xfmedia_settings_get_int("/xfmedia/general/height_large");
        if(h_l == -1) {
            gtk_window_get_size(GTK_WINDOW(topwin), &w, &h_l);
            xfmedia_settings_set_int("/xfmedia/general/height_large", h_l);
            xfmedia_settings_set_int("/xfmedia/general/width", w);
        }
    } else if(h_s == -1) {
        gtk_window_get_size(GTK_WINDOW(topwin), &w, &h_s);
        xfmedia_settings_set_int("/xfmedia/general/height_small", h_s);
        xfmedia_settings_set_int("/xfmedia/general/width", w);
    }
    
    if(xfmedia_settings_get_bool("/xfmedia/general/window_sticky"))
        gtk_window_stick(GTK_WINDOW(topwin));
    gtk_window_set_keep_above(GTK_WINDOW(topwin),
            xfmedia_settings_get_bool("/xfmedia/general/window_above"));
    
    mwin->pos_id = g_timeout_add(500, xfmedia_position_cb, mwin);
    
    return mwin;
}

void
xfmedia_mainwin_set_infobar_type(XfmediaMainwin *mwin, XfmediaInfobarType type)
{
    g_return_if_fail(mwin && type < XFMEDIA_INFOBAR_N_TYPES);
    
    switch(type) {
        case XFMEDIA_INFOBAR_SONG_LABEL:
            if(GTK_WIDGET_VISIBLE(mwin->progress_box))
                gtk_widget_hide(mwin->progress_box);
            if(GTK_WIDGET_VISIBLE(mwin->ui_message_label))
                gtk_widget_hide(mwin->ui_message_label);
            
            if(!GTK_WIDGET_VISIBLE(mwin->song_label))
                gtk_widget_show(mwin->song_label);
            break;
        
        case XFMEDIA_INFOBAR_PROGRESS_MESSAGE:
            if(GTK_WIDGET_VISIBLE(mwin->song_label))
                gtk_widget_hide(mwin->song_label);
            if(GTK_WIDGET_VISIBLE(mwin->ui_message_label))
                gtk_widget_hide(mwin->ui_message_label);
            
            if(!GTK_WIDGET_VISIBLE(mwin->progress_box))
                gtk_widget_show(mwin->progress_box);
            break;
        
        case XFMEDIA_INFOBAR_UI_MESSAGE:
            if(GTK_WIDGET_VISIBLE(mwin->song_label))
                gtk_widget_hide(mwin->song_label);
            if(GTK_WIDGET_VISIBLE(mwin->progress_box))
                gtk_widget_hide(mwin->progress_box);
            
            if(!GTK_WIDGET_VISIBLE(mwin->ui_message_label))
                gtk_widget_show(mwin->ui_message_label);
            break;
        
        default:
            break;
    }
}

gboolean
xfmedia_mainwin_get_file_info(XfmediaMainwin *mwin, const gchar *filename,
        gchar **title, gint *length)
{
    XfmediaTagInfo *tag_info;
    gchar *title_tmp;
    
    g_return_val_if_fail(title && length, FALSE);
    
    if(!filename)
        filename = xfmedia_xine_get_current_mrl(mwin->xfx);
    
    tag_info = xfmedia_tag_open(mwin, filename);
    g_return_val_if_fail(tag_info, FALSE);
    
    title_tmp = xfmedia_tag_to_string(tag_info,
            xfmedia_settings_get_string("/xfmedia/playlist/playlist_format"));
    if(title_tmp) {
        *title = xfmedia_brutal_utf8_validate(title_tmp);
        if(*title)
            g_free(title_tmp);
        else
            *title = title_tmp;
    }
    
    *length = xfmedia_tag_get_stream_length(tag_info);
    
    xfmedia_tag_close(tag_info);
    
    return TRUE;
}

void
xfmedia_mainwin_set_time_label(XfmediaMainwin *mwin, gint time,
        gboolean is_remaining, const gchar *extra_pango_markup)
{
    gchar time_lbl[128];
    gint h, m, s;
    
    if(time >= 3600) {
        h = time / 3600;
        m = (time % 3600) / 60;
        s = (time % 3600) % 60;
        
        g_snprintf(time_lbl, 128,
                   "<span font_family=\"monospace\" size=\"large\" weight=\"bold\" %s>%c%02d:%02d:%02d</span>",
                   extra_pango_markup ? extra_pango_markup : "",
                   is_remaining ? '-' : ' ', h, m, s);
    } else {
        m = time / 60;
        s = time % 60;
        
        g_snprintf(time_lbl, 128,
                   "<span font_family=\"monospace\" size=\"large\" weight=\"bold\" %s>%c%02d:%02d</span>",
                   extra_pango_markup ? extra_pango_markup : "",
                   is_remaining ? '-' : ' ', m, s);
    }
    gtk_label_set_markup(GTK_LABEL(mwin->time_label), time_lbl);
    
    if(time >= 3600) {
        g_snprintf(time_lbl, 128,
                   "<span font_family=\"monospace\" %s>%c%02d:%02d:%02d</span>",
                   extra_pango_markup ? extra_pango_markup : "",
                   is_remaining ? '-' : ' ', h, m, s);
    } else {
        g_snprintf(time_lbl, 128,
                   "<span font_family=\"monospace\" %s>%c%02d:%02d</span>",
                   extra_pango_markup ? extra_pango_markup : "",
                   is_remaining ? '-' : ' ', m, s);
    }
    gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label), time_lbl);
}

static gint
get_base_text_size()
{
    GtkWidget *dummy;
    gint base_size = -1;
    
    dummy = gtk_invisible_new();
    gtk_widget_realize(dummy);
    
    if(dummy->style)
        base_size = pango_font_description_get_size(dummy->style->font_desc);
    
    gtk_widget_destroy(dummy);
    
    return base_size;
}

void
xfmedia_mainwin_set_song_label(XfmediaMainwin *mwin, const gchar *str)
{
    gint base_size;
    gchar *titlebar_str;
    
    xfmedia_mainwin_set_infobar_type(mwin, XFMEDIA_INFOBAR_SONG_LABEL);
    
    g_return_if_fail(mwin && str);
    
    base_size = get_base_text_size();
    if(base_size != -1) {
        PangoFontDescription *font_desc = pango_font_description_new();
        pango_font_description_set_size(font_desc, base_size*PANGO_SCALE_LARGE);
        gtk_widget_modify_font(mwin->song_label, font_desc);
        pango_font_description_free(font_desc);
        gtk_label_set_use_markup(GTK_LABEL(mwin->song_label), FALSE);
        gtk_label_set_text(GTK_LABEL(mwin->song_label), str);
    } else {
        gchar *str_esc, *str_markup;
        str_esc = g_markup_escape_text(str, strlen(str));
        str_markup = g_strdup_printf("<span size=\"large\">%s</span>", str_esc);
        g_free(str_esc);
        gtk_label_set_use_markup(GTK_LABEL(mwin->song_label), TRUE);
        gtk_label_set_markup(GTK_LABEL(mwin->song_label), str_markup);
        g_free(str_markup);
    }
    
    titlebar_str = g_strdup_printf("Xfmedia: %s", str);
    gtk_window_set_title(GTK_WINDOW(mwin->window), titlebar_str);
    g_free(titlebar_str);
    
    if(xfmedia_settings_get_bool("/xfmedia/general/show_tray_icon")) {
        gchar *str_tt = g_strdup_printf("%s: %s", _("Playing"), str);
        xfmedia_tray_icon_set_tooltip(mwin->tray_icon, str_tt);
        g_free(str_tt);
    }
}

static GtkWidget *
tray_icon_menu_create(XfmediaMainwin *mwin)
{
    GtkWidget *menu, *mi;
    gint iw, ih;
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &iw, &ih);
    
    menu = gtk_menu_new();
    gtk_widget_show(menu);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("P_revious Track"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_prev_cb), mwin);
    
    mwin->tray_icon_playpause_btn = mi =
            gtk_image_menu_item_new_with_mnemonic(_("_Play"));
    mwin->tray_icon_play_img = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY,
            GTK_ICON_SIZE_MENU);
    g_object_ref(G_OBJECT(mwin->tray_icon_play_img));
    if(xfmedia_xine_get_param(mwin->xfx, XINE_PARAM_SPEED) != XINE_SPEED_NORMAL
            || xfmedia_xine_get_status(mwin->xfx) != XINE_STATUS_PLAY)
    {
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
                mwin->tray_icon_play_img);
    }
    mwin->tray_icon_pause_img = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PAUSE,
            GTK_ICON_SIZE_MENU);
    g_object_ref(G_OBJECT(mwin->tray_icon_pause_img));
    if(xfmedia_xine_get_param(mwin->xfx, XINE_PARAM_SPEED) == XINE_SPEED_NORMAL
            && xfmedia_xine_get_status(mwin->xfx) == XINE_STATUS_PLAY)
    {
        gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(mi))),
                _("_Pause"));
        gtk_label_set_use_underline(GTK_LABEL(gtk_bin_get_child(GTK_BIN(mi))),
                TRUE);
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
                mwin->tray_icon_pause_img);
    }
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playpause_cb), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Stop"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_stop_cb), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Next Track"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_NEXT, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_next_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    mi = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
            G_CALLBACK(gtk_widget_destroy), mwin->window);
    
    return menu;
}

void
xfmedia_mainwin_activate_systray(XfmediaMainwin *mwin)
{
    GdkScreen *gscreen = gtk_widget_get_screen(mwin->window);
    
    if(!mwin->tray_icon_menu)
        mwin->tray_icon_menu = tray_icon_menu_create(mwin);
    
    mwin->tray_icon = xfmedia_tray_icon_create(gscreen, mwin,
            GTK_MENU(mwin->tray_icon_menu));
}

void
xfmedia_mainwin_deactivate_systray(XfmediaMainwin *mwin)
{
    g_return_if_fail(mwin->tray_icon);
    
    xfmedia_tray_icon_destroy(mwin->tray_icon);
    mwin->tray_icon = NULL;
}

void
xfmedia_mainwin_connect_slider_callback(XfmediaMainwin *mwin)
{
    g_signal_handlers_unblock_by_func(G_OBJECT(mwin->position_slider),
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
    g_signal_handlers_unblock_by_func(G_OBJECT(mwin->fs_position_slider),
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
}

void
xfmedia_mainwin_disconnect_slider_callback(XfmediaMainwin *mwin)
{
    g_signal_handlers_block_by_func(G_OBJECT(mwin->position_slider),
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
    g_signal_handlers_block_by_func(G_OBJECT(mwin->fs_position_slider),
            G_CALLBACK(xfmedia_position_slider_cb), mwin);
}

void
xfmedia_mainwin_set_play_buttons(XfmediaMainwin *mwin)
{
    gtk_container_remove(GTK_CONTAINER(mwin->playpause_btn),
            gtk_bin_get_child(GTK_BIN(mwin->playpause_btn)));
    gtk_container_add(GTK_CONTAINER(mwin->playpause_btn),
            mwin->play_btn_img);
    gtk_container_remove(GTK_CONTAINER(mwin->fs_playpause_btn),
            gtk_bin_get_child(GTK_BIN(mwin->fs_playpause_btn)));
    gtk_container_add(GTK_CONTAINER(mwin->fs_playpause_btn),
            mwin->fs_play_btn_img);
    if(mwin->tray_icon_playpause_btn && mwin->tray_icon_play_img) {
        GtkWidget *lbl;
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mwin->tray_icon_playpause_btn),
                mwin->tray_icon_play_img);
        lbl = gtk_bin_get_child(GTK_BIN(mwin->tray_icon_playpause_btn));
        gtk_label_set_text(GTK_LABEL(lbl), _("_Play"));
        gtk_label_set_use_underline(GTK_LABEL(lbl), TRUE);
    }
}

void
xfmedia_mainwin_set_pause_buttons(XfmediaMainwin *mwin)
{
    gtk_container_remove(GTK_CONTAINER(mwin->playpause_btn),
            gtk_bin_get_child(GTK_BIN(mwin->playpause_btn)));
    gtk_container_add(GTK_CONTAINER(mwin->playpause_btn),
            mwin->pause_btn_img);
    gtk_container_remove(GTK_CONTAINER(mwin->fs_playpause_btn),
            gtk_bin_get_child(GTK_BIN(mwin->fs_playpause_btn)));
    gtk_container_add(GTK_CONTAINER(mwin->fs_playpause_btn),
            mwin->fs_pause_btn_img);
    if(mwin->tray_icon_playpause_btn && mwin->tray_icon_pause_img) {
        GtkWidget *lbl;
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mwin->tray_icon_playpause_btn),
                mwin->tray_icon_pause_img);
        lbl = gtk_bin_get_child(GTK_BIN(mwin->tray_icon_playpause_btn));
        gtk_label_set_text(GTK_LABEL(lbl), _("_Pause"));
        gtk_label_set_use_underline(GTK_LABEL(lbl), TRUE);
    }
}

void
xfmedia_mainwin_destroy(XfmediaMainwin *mwin)
{    
    xfmedia_keybindings_cleanup();
    
    if(mwin->xfx)
        xfmedia_xine_stop(mwin->xfx);
    
    if(mwin->pos_id) {
        g_source_remove(mwin->pos_id);
        mwin->pos_id = 0;
    }
    
    if(mwin->tooltips) {
        gtk_object_sink(GTK_OBJECT(mwin->tooltips));
        mwin->tooltips = NULL;
    }
    
    if(mwin->shuffle_history) {
        gint i, tot = g_queue_get_length(mwin->shuffle_history);
        for(i = 0; i < tot; i++)
            xfmedia_playlist_entry_ref_destroy((XfmediaPlaylistEntryRef *)g_queue_peek_nth(mwin->shuffle_history, i));
        g_queue_free(mwin->shuffle_history);
        mwin->shuffle_history = NULL;
    }
    
    DBG("closing extra stream and a/v drivers");
    if(mwin->extra_stream) {
        g_mutex_lock(mwin->extra_xine_mx);
        xine_close(mwin->extra_stream);
        xine_dispose(mwin->extra_stream);
        xine_close_audio_driver(xfmedia_xine_get_raw_xine_engine(mwin->xfx),
                mwin->extra_ap);
        xine_close_video_driver(xfmedia_xine_get_raw_xine_engine(mwin->xfx),
                mwin->extra_vp);
        mwin->extra_stream = NULL;
        mwin->extra_ap = NULL;
        mwin->extra_vp = NULL;
        g_mutex_unlock(mwin->extra_xine_mx);
        g_mutex_free(mwin->extra_xine_mx);
        mwin->extra_xine_mx = NULL;
    }
    
    DBG("killing tray icon");
    if(mwin->tray_icon) {
        xfmedia_tray_icon_destroy(mwin->tray_icon);
        mwin->tray_icon = NULL;
    }
    
    if(mwin->fs_controls) {
        gtk_widget_destroy(mwin->fs_controls);
        mwin->fs_controls = NULL;
    }
    
    if(mwin->autohide_window) {
        gtk_widget_destroy(mwin->autohide_window);
        mwin->autohide_window = NULL;
    }
    
    if(mwin->add_menu) {
        gtk_widget_destroy(mwin->add_menu);
        mwin->add_menu = NULL;
    }
    
    if(mwin->playlist_q) {
        /* destroyed with parent even if not mapped? */
        /*gtk_widget_destroy(GTK_WIDGET(mwin->playlist_q));
        mwin->playlist_q = NULL;*/
    }
    
    g_free(mwin);
    mwin = NULL;
}

void
xfmedia_mainwin_show(XfmediaMainwin *mwin)
{
    if(!GTK_WIDGET_VISIBLE(mwin->window)) {
        gtk_widget_show(mwin->window);
        gtk_window_move(GTK_WINDOW(mwin->window),
                xfmedia_settings_get_int("/xfmedia/general/pos_x"),
                xfmedia_settings_get_int("/xfmedia/general/pos_y"));
    } else if(gdk_window_get_state(mwin->window->window) & GDK_WINDOW_STATE_ICONIFIED)
        gtk_window_deiconify(GTK_WINDOW(mwin->window));
    
    if(GTK_WIDGET_VISIBLE(mwin->autohide_window))
        gtk_widget_hide(mwin->autohide_window);
    
    gtk_window_set_skip_pager_hint(GTK_WINDOW(mwin->window), FALSE);
    gtk_window_set_skip_taskbar_hint(GTK_WINDOW(mwin->window), FALSE);
}

static gboolean
xfmedia_mainwin_try_play_http_playlist(XfmediaMainwin *mwin, const gchar *uri,
        gint idx_ref)
{
    gboolean ret = FALSE;
    
    if(!g_ascii_strncasecmp(uri, "http://", 7)) {
        /* we may have some kind of http playlist; try to grab it and see
         * if it's something we can parse  */
        gchar buf[256], mime_type[256], *cachefile = NULL;
        gint bin;
        
        memset(mime_type, 0, sizeof(buf));
        
        DBG("trying to peek at file");
        
        /* disable cache lookups for now because for some reason it causes
         * a double-free() crash */
        /*if(!(cachefile=xfmedia_playlists_cache_find(uri))) {*/
            bin = http_peek(uri, sizeof(buf), buf, mime_type);
            if(bin > 0) {
                DBG("got file with mime type %s", mime_type);
                if(!strcmp(mime_type, "video/x-ms-asf")) {
                    gchar *asx = http_download(uri, &bin);
                    if(asx && bin > 0) {
                        cachefile = xfmedia_playlists_cache_save(uri, asx,
                                "asx");
                        g_free(asx);
                    }
                } else if(!strcmp(mime_type, "audio/x-mpegurl")) {
                    gchar *m3u = http_download(uri, &bin);
                    if(m3u && bin > 0) {
                        cachefile = xfmedia_playlists_cache_save(uri, m3u,
                                "m3u");
                        g_free(m3u);
                    }
                } else if(!strcmp(mime_type, "audio/x-scpls")) {
                    gchar *pls = http_download(uri, &bin);
                    if(pls && bin > 0) {
                        cachefile = xfmedia_playlists_cache_save(uri, pls,
                                "pls");
                        g_free(pls);
                    }
                } else if(!strcmp(mime_type, "audio/x-ms-wax")) {
                    gchar *wax = http_download(uri, &bin);
                    if(wax && bin > 0) {
                        cachefile = xfmedia_playlists_cache_save(uri, wax,
                                "wax");
                        g_free(wax);
                    }
                }
            }
        /*}*/
        
        if(cachefile) {
            gchar *entry1, *title = NULL;
            gint length = -1;
            
            DBG("trying to play cached playlist %s", cachefile);
            
            entry1 = xfmedia_playlists_get_first_entry(cachefile, &title,
                    &length);
            if(entry1) {
                if(idx_ref >= 0 && title) {
                    gchar *title_utf8 = xfmedia_brutal_utf8_validate(title);
                    xfmedia_playlist_modify_entry(mwin->plist, idx_ref,
                            title_utf8 ? title_utf8 : title, length, NULL);
                    xfmedia_playlist_set_metadata_loaded(mwin->plist,
                            idx_ref, TRUE);
                    g_free(title_utf8);
                }
                DBG("trying to play playlist entry %s", entry1);
                ret = xfmedia_mainwin_play_uri(mwin, entry1, -1);
                g_free(entry1);
            }
            g_free(cachefile);
        }
    }
    
    return ret;
}

gboolean
xfmedia_mainwin_play_uri(XfmediaMainwin *mwin, const gchar *uri, gint idx_ref)
{
    gboolean ret = FALSE;
    
    xfmedia_xine_stop(mwin->xfx);   
    DBG("about to call xfmedia_xine_open()"); 
    if(xfmedia_xine_open(mwin->xfx, uri)) {
        gboolean was_visible = (GTK_WIDGET_VISIBLE(mwin->video_window)
                || xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window)));
        gboolean has_video = xfmedia_xine_get_stream_info(mwin->xfx,
                XINE_STREAM_INFO_HAS_VIDEO);
        gboolean needs_vwin_show = FALSE;
        
        DBG("  xfmedia_xine_open() returned");
        
        xfmedia_plugins_forward_signal_VOID("stream-loaded");
        
        has_video = xfmedia_xine_get_stream_info(mwin->xfx,
                XINE_STREAM_INFO_HAS_VIDEO);
        
        if(!xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis")
                && was_visible && !has_video)
        {
            if(GTK_WIDGET_VISIBLE(mwin->fs_controls))
                gtk_widget_hide(mwin->fs_controls);
            gtk_widget_hide(mwin->video_window);
        } else if(has_video) {
            if(!was_visible)
                needs_vwin_show = TRUE;
            xfmedia_xine_set_vis(mwin->xfx, NULL);
        } else if(xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis")
                && !has_video)
        {
            needs_vwin_show = TRUE;
            xfmedia_xine_set_vis(mwin->xfx,
                    xfmedia_settings_get_string("/xfmedia/general/selected_vis"));
        }
        
        DBG("dealt with video window");
        
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 0);
        if(!xfmedia_xine_play(mwin->xfx, 0, 0)) {
            DBG("xfmedia_xine_play() failed");
            ret = xfmedia_mainwin_try_play_http_playlist(mwin, uri, idx_ref);
            if(!ret) {
                if(!was_visible)
                    gtk_widget_hide(mwin->video_window);
                xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
            }
        } else {
            gchar *name = NULL;
            gint song_len = -1;
            
            if(!xfmedia_mainwin_get_file_info(mwin, NULL, &name, &song_len) || !name)
                name = xfmedia_filename_to_name(uri);
            xfmedia_mainwin_set_song_label(mwin, name);
            g_free(name);
            
            if(has_video) {
                gint width, height;
                
                width = xfmedia_xine_get_stream_info(mwin->xfx,
                        XINE_STREAM_INFO_VIDEO_WIDTH);
                height = xfmedia_xine_get_stream_info(mwin->xfx,
                        XINE_STREAM_INFO_VIDEO_HEIGHT);
                if(width < 1 || height < 1) {
                    width = 480;
                    height = 360;
                }
                
                xfmedia_video_window_set_aspect(XFMEDIA_VIDEO_WINDOW(mwin->video_window),
                                                width, height);
                if(!xfmedia_video_window_has_fixed_size(XFMEDIA_VIDEO_WINDOW(mwin->video_window))
                   && !xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window))
                   && needs_vwin_show)
                {
                    gtk_window_resize(GTK_WINDOW(mwin->video_window),
                                      width, height);                        
                }
            }
            
            if(needs_vwin_show
               || xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis"))
            {
                gtk_widget_show(mwin->video_window);
            }
            
            xfmedia_mainwin_disconnect_slider_callback(mwin);
            gtk_range_set_value(GTK_RANGE(mwin->position_slider), 0.0);
            gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), 0.0);
            xfmedia_mainwin_connect_slider_callback(mwin);
            
            if(xfmedia_xine_get_stream_info(mwin->xfx, XINE_STREAM_INFO_SEEKABLE)) {
                gtk_widget_set_sensitive(mwin->position_slider, TRUE);
                gtk_widget_set_sensitive(mwin->fs_position_slider, TRUE);
            } else {
                gtk_widget_set_sensitive(mwin->position_slider, FALSE);
                gtk_widget_set_sensitive(mwin->fs_position_slider, FALSE);
            }
            
            mwin->cur_playing_length = song_len;
            if(song_len <= 0)
                song_len = 1;
            
            gtk_range_set_range(GTK_RANGE(mwin->position_slider), 0.0,
                    song_len>0 ? (gdouble)song_len : 1.0);
            gtk_range_set_range(GTK_RANGE(mwin->fs_position_slider), 0.0,
                    song_len>0 ? (gdouble)song_len : 1.0);
            
            xfmedia_mainwin_set_pause_buttons(mwin);
            
            xfmedia_plugins_forward_signal_VOID("stream-started");
            ret = TRUE;
        }
    } else {
        DBG("xine_open() failed, checking URI");
        return xfmedia_mainwin_try_play_http_playlist(mwin, uri, idx_ref);
    }
    
    return ret;
}

static gint
mainwin_get_next_index(XfmediaMainwin *mwin, gint last_index)
{
    gint tot_items, next_id, attempts = 0, cutoff;
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")) {
        tot_items = xfmedia_playlist_get_n_entries(mwin->plist);
        cutoff = tot_items * 0.75;
        
        if(tot_items == 0) {
            /* this happened once; not sure why */
            g_warning("tot_items == 0, checking again");
            tot_items = xfmedia_playlist_get_n_entries(mwin->plist);
            if(tot_items == 0) {
                g_critical("tot_items STILL == 0");
                tot_items = 1;
            } else
                g_warning("tot_items set properly on second try (%d)", tot_items);
        }
        
        do {
            if(attempts > 0)
                DBG("shuffle collision!  trying again (attempt %d)", attempts);
            
#if HAVE_SRANDOM
            next_id = random() % tot_items;
#else
            next_id = rand() % tot_items;
#endif
            if(++attempts > cutoff) {
                xfmedia_playlist_shuffle_unset_all(mwin->plist);
                attempts = 0;
            }
        } while(!xfmedia_playlist_shuffle_check(mwin->plist, next_id));
    } else if(last_index == -1 && mwin->cur_playing) {
        tot_items = xfmedia_playlist_get_n_entries(mwin->plist);
        if(tot_items == 0)
            return -1;
        last_index = xfmedia_playlist_entry_ref_get_index(mwin->cur_playing);
        if(last_index >= tot_items - 1) {
            /* this means we're at the end of the list (i hope) */
            if(xfmedia_settings_get_bool("/xfmedia/playlist/repeat"))
                next_id = 0;
            else {
                xfmedia_xine_stop(mwin->xfx);
                if(xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window))) {
                    xfmedia_video_window_set_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window),
                                                        FALSE);
                }
                if(GTK_WIDGET_VISIBLE(mwin->video_window))
                    gtk_widget_hide(mwin->video_window);
                next_id = -1;
            }
        } else
            next_id = last_index + 1;
    } else
        next_id = 0;
    
    return next_id;
}

gboolean
xfmedia_mainwin_play_file_at_index(XfmediaMainwin *mwin, gint idx)
{
    gchar *filename;
    gboolean ret = FALSE;
    guint idx_start = 0, idx_end = 0;
    
    if(idx == -1) {
        idx = mainwin_get_next_index(mwin, -1);
        if(idx == -1)
            return FALSE;
    }
    
    if(mwin->cur_playing) {
        xfmedia_playlist_entry_ref_destroy(mwin->cur_playing);
        mwin->cur_playing = NULL;
        mwin->cur_playing_length = 0;
    }
    
    if(xfmedia_playlist_get(mwin->plist, idx, NULL, NULL, &filename) && filename) {
        if(!xfmedia_playlist_get_metadata_loaded(mwin->plist, idx)) {
            gchar *title = NULL;
            gint length = 0;
            
            if(xfmedia_mainwin_get_file_info(mwin, filename, &title, &length)) {
                if((title && !*title) || !title) {
                    g_free(title);
                    title = xfmedia_filename_to_name(filename);
                }
                xfmedia_playlist_modify_entry(mwin->plist, idx, title, length,
                        filename);
                g_free(title);
                xfmedia_playlist_set_metadata_loaded(mwin->plist, idx, TRUE);
            }
        }
        
        if(xfmedia_playlist_get_visible_range(mwin->plist, &idx_start, &idx_end)) {
            if(idx_start > idx || idx_end < idx)
                xfmedia_playlist_scroll_to_index(mwin->plist, idx, 0.5);
        }
        mwin->cur_playing = xfmedia_playlist_entry_ref_new_for_index(mwin->plist, idx);
        xfmedia_playlist_set_bold_entry(mwin->plist, mwin->cur_playing);
        xfmedia_settings_set_int("/xfmedia/playlist/playlist_position", idx);
        xfmedia_playlist_shuffle_check_set(mwin->plist, idx);
        
        if(xfmedia_mainwin_play_uri(mwin, filename, idx))
            ret = TRUE;
        
        g_free(filename);
    }
    
    return ret;
}

gboolean
xfmedia_mainwin_play_file_at_index_persistent(XfmediaMainwin *mwin, gint idx)
{
    gboolean done = FALSE, played = FALSE;
    gboolean auto_advance = xfmedia_settings_get_bool("/xfmedia/playlist/auto_advance");
    
    do {
        played = xfmedia_mainwin_play_file_at_index(mwin, idx);
        if(played || !auto_advance)
            done = TRUE;
        
        if(!done) {
            idx = mainwin_get_next_index(mwin, idx);
            if(idx == -1)
                done = TRUE;
        }
    } while(!done);
    
    return played;
}

static inline void
mainwin_set_tooltip_playing(XfmediaMainwin *mwin)
{
    if(mwin->cur_playing && xfmedia_settings_get_bool("/xfmedia/general/show_tray_icon")) {
        gint idx = xfmedia_playlist_entry_ref_get_index(mwin->cur_playing);
        gchar *title = NULL;
        
        if(idx != -1
                && xfmedia_playlist_get(mwin->plist, idx, &title, NULL, NULL)
                && title)
        {
            gchar *str = g_strdup_printf("%s: %s", _("Playing"), title);
            xfmedia_tray_icon_set_tooltip(mwin->tray_icon, str);
            g_free(str);
        }
    }
}

void
xfmedia_mainwin_toggle_playpause(XfmediaMainwin *mwin)
{
    gint speed, status;
    
    speed = xfmedia_xine_get_param(mwin->xfx, XINE_PARAM_SPEED);
    status = xfmedia_xine_get_status(mwin->xfx);
    
    if(status == XINE_STATUS_PLAY && speed != XINE_SPEED_PAUSE) {
        xfmedia_mainwin_set_play_buttons(mwin);
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_SPEED, XINE_SPEED_PAUSE);
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
        xfmedia_plugins_forward_signal_VOID("speed-changed");
        if(xfmedia_settings_get_bool("/xfmedia/general/show_tray_icon"))
            xfmedia_tray_icon_set_tooltip(mwin->tray_icon, _("Paused"));
    } else if(status == XINE_STATUS_STOP || status == XINE_STATUS_IDLE) {
        if(!xfmedia_xine_play(mwin->xfx, 0, 0)) {
            gint idx = -1;
            
            if(mwin->cur_playing)
            	idx = xfmedia_playlist_entry_ref_get_index(mwin->cur_playing);
            if(idx < 0)
                idx = xfmedia_playlist_get_selected(mwin->plist);
            if(idx < 0)
                idx = 0;
            
            xfmedia_mainwin_play_file_at_index(mwin, idx);
        } else {
            if(xfmedia_xine_get_stream_info(mwin->xfx, XINE_STREAM_INFO_HAS_VIDEO)
                    || xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis"))
            {
                gint width, height;
                
                width = xfmedia_xine_get_stream_info(mwin->xfx,
                        XINE_STREAM_INFO_VIDEO_WIDTH);
                height = xfmedia_xine_get_stream_info(mwin->xfx,
                        XINE_STREAM_INFO_VIDEO_HEIGHT);
                if(width < 1 || height < 1) {
                    width = 480;
                    height = 360;
                }
                
                xfmedia_video_window_set_aspect(XFMEDIA_VIDEO_WINDOW(mwin->video_window),
                                                width, height);
                
                if(!xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window))
                        && !GTK_WIDGET_VISIBLE(mwin->video_window))
                {
                    gtk_widget_show(mwin->video_window);
                }
                
                if(!xfmedia_video_window_has_fixed_size(XFMEDIA_VIDEO_WINDOW(mwin->video_window))) {
                    gtk_window_resize(GTK_WINDOW(mwin->video_window),
                                      width, height);
                }
                
                mainwin_set_tooltip_playing(mwin);
            }
        }
    } else if(status == XINE_STATUS_PLAY && speed == XINE_SPEED_PAUSE) {
        xfmedia_mainwin_set_pause_buttons(mwin);
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 0);
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
        xfmedia_plugins_forward_signal_VOID("speed-changed");
        mainwin_set_tooltip_playing(mwin);
    }
}

void
xfmedia_mainwin_stop(XfmediaMainwin *mwin)
{
    gint speed, status;
    
    speed = xfmedia_xine_get_param(mwin->xfx, XINE_PARAM_SPEED);
    status = xfmedia_xine_get_status(mwin->xfx);
    
    if(status == XINE_STATUS_PLAY && speed != XINE_SPEED_PAUSE)
        xfmedia_mainwin_set_play_buttons(mwin);
    
    if(status == XINE_STATUS_PLAY) {
        xfmedia_xine_stop(mwin->xfx);
        
        xfmedia_mainwin_disconnect_slider_callback(mwin);
        gtk_range_set_value(GTK_RANGE(mwin->position_slider), 0.0);
        gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), 0.0);
        xfmedia_mainwin_connect_slider_callback(mwin);
        
        gtk_label_set_markup(GTK_LABEL(mwin->time_label),
                "<span font_family=\"monospace\" size=\"large\" weight=\"bold\"> 00:00</span>");
        gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label),
                "<span font_family=\"monospace\"> 00:00</span>");
        
        if(GTK_WIDGET_VISIBLE(mwin->video_window))
            gtk_widget_hide(mwin->video_window);
        if(xfmedia_video_window_get_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window))) {
            if(GTK_WIDGET_VISIBLE(mwin->fs_controls))
                gtk_widget_hide(mwin->fs_controls);
            xfmedia_video_window_set_fullscreen(XFMEDIA_VIDEO_WINDOW(mwin->video_window), FALSE);
        }
        xfmedia_xine_set_param(mwin->xfx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
    }
    
    if(xfmedia_settings_get_bool("/xfmedia/general/show_tray_icon"))
        xfmedia_tray_icon_set_tooltip(mwin->tray_icon, _("Stopped"));
}

void
xfmedia_mainwin_next(XfmediaMainwin *mwin)
{
    gint next_index = -1;
    gboolean use_playlist_q  = xfmedia_playlist_queue_n_entries(mwin->playlist_q);
    gboolean ret = FALSE;
    
    if(mwin->cur_playing
       && xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")
       && !use_playlist_q )
    {
        DBG("putting item on shuffle history");
        g_queue_push_head(mwin->shuffle_history,
                xfmedia_playlist_entry_ref_copy(mwin->cur_playing));
    }
    
    if(use_playlist_q) {
        XfmediaPlaylistEntryRef *ref;
        gint idx;
        do {
            ref = xfmedia_playlist_queue_pop_entry(mwin->playlist_q);
            if(ref && (idx = xfmedia_playlist_entry_ref_get_index(ref)) >= 0) {
                ret = xfmedia_mainwin_play_file_at_index(mwin, idx);
                xfmedia_playlist_entry_ref_destroy(ref);
            } else {
                if(ref)
                    xfmedia_playlist_entry_ref_destroy(ref);
                break;
            }
        } while(!ret);
    }
    
    if(!ret)
        xfmedia_mainwin_play_file_at_index_persistent(mwin, next_index);
}

void
xfmedia_mainwin_prev(XfmediaMainwin *mwin)
{
    gint tot_items, idx = -1;
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")) {
        XfmediaPlaylistEntryRef *ref = NULL;
        gboolean playing = FALSE;
        
        if(mwin->cur_playing) {
        	idx = xfmedia_playlist_entry_ref_get_index(mwin->cur_playing);
			if(idx >= 0)
                xfmedia_playlist_shuffle_unset(mwin->plist, idx);
        }
        
        do {
            if(idx != -1)
                xfmedia_playlist_shuffle_unset(mwin->plist, idx);
            ref = g_queue_pop_head(mwin->shuffle_history);
            DBG("got ref from shuffle history %p", ref);
            if(ref) {
                idx = xfmedia_playlist_entry_ref_get_index(ref);
                xfmedia_playlist_entry_ref_destroy(ref);
            } else {
                tot_items = xfmedia_playlist_get_n_entries(mwin->plist);
#if HAVE_SRANDOM
                idx = random() % tot_items;
#else
                idx = rand() % tot_items;
#endif
            }
        } while(idx != -1
                && !(playing=xfmedia_mainwin_play_file_at_index(mwin, idx)));
        
        if(!playing)
            xfmedia_mainwin_play_file_at_index_persistent(mwin, -1);
    } else if(mwin->cur_playing
            && xfmedia_playlist_entry_ref_valid(mwin->cur_playing))
    {
        idx = xfmedia_playlist_entry_ref_get_index(mwin->cur_playing);
        do {
            if(idx == 0 && xfmedia_settings_get_bool("/xfmedia/playlist/repeat")) {
                tot_items = xfmedia_playlist_get_n_entries(mwin->plist);
                idx = tot_items - 1;
            } else if(idx == -1)
                idx = 0;
            else
                idx--;
        } while(!xfmedia_mainwin_play_file_at_index(mwin, idx));
    }
}

void
xfmedia_mainwin_enable_autohide(XfmediaMainwin *mwin)
{
    g_return_if_fail(mwin && mwin->temp_disable_autohide > 0);
    if(--mwin->temp_disable_autohide == 0) {
        mwin->autohide_timer = g_timeout_add(AUTOHIDE_TIMEOUT,
                                             (GSourceFunc)xfmedia_mainwin_check_do_autohide,
                                             mwin);
    }
}

void
xfmedia_mainwin_disable_autohide(XfmediaMainwin *mwin)
{
    g_return_if_fail(mwin);
    mwin->temp_disable_autohide++;
}

void
xfmedia_mainwin_toggle_playlist_queue(XfmediaMainwin *mwin)
{
    if(!GTK_WIDGET_VISIBLE(GTK_WIDGET(mwin->playlist_q))) {
        gint x = xfmedia_settings_get_int("/xfmedia/playlist_queue/pos_x");
        gint y = xfmedia_settings_get_int("/xfmedia/playlist_queue/pos_y");
        
        if(x != -1 && y != -1)
            gtk_window_move(GTK_WINDOW(mwin->playlist_q), x, y);
        gtk_widget_show(GTK_WIDGET(mwin->playlist_q));
    } else
        gtk_widget_hide(GTK_WIDGET(mwin->playlist_q));
    
    g_signal_handlers_block_by_func(G_OBJECT(mwin->playlist_q_mi),
                                    xfmedia_mainwin_toggle_playlist_queue,
                                    mwin);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->playlist_q_mi),
                                   GTK_WIDGET_VISIBLE(GTK_WIDGET(mwin->playlist_q)));
    g_signal_handlers_unblock_by_func(G_OBJECT(mwin->playlist_q_mi),
                                    xfmedia_mainwin_toggle_playlist_queue,
                                    mwin);
}

void
xfmedia_mainwin_queue_selected(XfmediaMainwin *mwin)
{
    gint idx = xfmedia_playlist_get_selected(mwin->plist);
    XfmediaPlaylistEntryRef *ref;
    gchar *title_utf8 = NULL;
    gint length;
    
    if(idx < 0)
        return;
    
    xfmedia_playlist_get(mwin->plist, idx, &title_utf8, &length, NULL);
    ref = xfmedia_playlist_entry_ref_new_for_index(mwin->plist, idx);
    
    xfmedia_playlist_queue_push_entry(mwin->playlist_q, ref, title_utf8, length);
    g_free(title_utf8);
    xfmedia_playlist_entry_ref_destroy(ref);
}
